@@ -20,7 +20,7 @@
  *
  */
 
-/* $Id: ffe.c,v 1.51 2007-05-30 07:32:48 timo Exp $ */
+/* $Id: ffe.c,v 1.80 2008-07-11 14:13:02 timo Exp $ */
 
 #include "ffe.h"
 
@@ -44,7 +44,7 @@ static char *program = "ffe";
 #ifdef PACKAGE_VERSION
 static char *version = PACKAGE_VERSION;
 #else
-static char *version = "0.2.2";
+static char *version = "0.2.5";
 #endif
 
 #ifdef HOST
@@ -59,7 +59,7 @@ static char *email_address = PACKAGE_BUGREPORT;
 static char *email_address = "tjsa@iki.fi";
 #endif
 
-static char short_opts[] = "c:s:o:p:f:e:r:l?Vav";
+static char short_opts[] = "c:s:o:p:f:e:r:l?VavdI";
 
 #ifdef HAVE_GETOPT_LONG
 static struct option long_opts[] = {
@@ -75,10 +75,20 @@ static struct option long_opts[] = {
     {"and",0,NULL,'a'},
     {"invert-match",0,NULL,'v'},
     {"replace",1,NULL,'r'},
+    {"debug",0,NULL,'d'},
+    {"info",0,NULL,'I'},
     {NULL,0,NULL,0}
 };
 #endif
 
+extern void print_info();
+
+extern struct input_file *files;
+
+int system_endianess = F_UNKNOWN_ENDIAN;
+int max_binary_record_length = 0;
+
+
 /* global rc-data */
 struct structure *structure = NULL;
 struct output *output = NULL;
@@ -88,8 +98,10 @@ struct replace *replace = NULL;
 struct field *const_field = NULL;
 
 /* output no marker */
-struct output dummy;
-struct output *no_output = &dummy;
+struct output dummy_no;
+struct output *no_output = &dummy_no;
+struct output dummy_raw;
+struct output *raw = &dummy_raw;
 
 void
 panic(char *msg,char *info,char *syserror)
@@ -110,6 +122,23 @@ panic(char *msg,char *info,char *syserror)
     exit(EXIT_FAILURE);
 }
 
+void
+problem(char *msg,char *info,char *syserror)
+{
+    if (info == NULL && syserror == NULL)
+    {
+        fprintf(stderr,"%s: %s\n",program,msg);
+    } else if(info != NULL && syserror == NULL)
+    {
+        fprintf(stderr,"%s: %s: %s\n",program,msg,info);
+    } else if(info != NULL && syserror != NULL)
+    {
+        fprintf(stderr,"%s: %s: %s; %s\n",program,msg,info,syserror);
+    } else if(info == NULL && syserror != NULL)
+    {
+        fprintf(stderr,"%s: %s; %s\n",program,msg,syserror);
+    }
+}
 
 
 char *
@@ -125,6 +154,16 @@ get_default_rc_name()
 
     result = NULL;
     home = getenv("HOME");
+#ifdef WIN32
+    if(home == NULL)
+    {
+        home = getenv("FFE_HOME");
+        if(home == NULL)
+        {
+            home = getenv("USERPROFILE");
+        }
+    }
+#endif
     if(home != NULL)
     {
             result = xmalloc(strlen(home) + strlen(file) + strlen(PATH_SEPARATOR_STRING) + 2);
@@ -163,6 +202,10 @@ help(FILE *stream)
     fprintf(stream,"\t\tAn invalid input line does not cause %s to abort.\n",program);
     fprintf(stream,"-r, --replace=FIELD=VALUE\n");
     fprintf(stream,"\t\tReplace FIELDs contents with VALUE in output.\n");
+    fprintf(stream,"-d, --debug\n");
+    fprintf(stream,"\t\tWrite invalid input lines to error log.\n");
+    fprintf(stream,"-I, --info\n");
+    fprintf(stream,"\t\tShow the structure information and exit.\n");
     fprintf(stream,"-?, --help\n");
     fprintf(stream,"\t\tDisplay this help and exit.\n");
     fprintf(stream,"-V, --version\n");
@@ -187,6 +230,10 @@ help(FILE *stream)
     fprintf(stream,"\t\tAn invalid input line does not cause %s to abort.\n",program);
     fprintf(stream,"-r FIELD=VALUE\n");
     fprintf(stream,"\t\tReplace FIELDs contents with VALUE in output.\n");
+    fprintf(stream,"-d\n");
+    fprintf(stream,"\t\tWrite invalid input lines to error log.\n");
+    fprintf(stream,"-I\n");
+    fprintf(stream,"\t\tShow the structure information and exit.\n");
     fprintf(stream,"-?\n");
     fprintf(stream,"\t\tDisplay this help and exit.\n");
     fprintf(stream,"-V\n");
@@ -220,6 +267,7 @@ search_output(char *name)
     struct output *o = output;
 
     if(strcmp(name,"no") == 0) return no_output;
+    if(strcmp(name,"raw") == 0) return raw;
 
     while(o != NULL)
     {
@@ -268,7 +316,7 @@ check_rc(char *use_output)
     struct record *r,*fr;
     struct field *f;
     struct lookup *l;
-    int several_records;
+    int several_records = 0;
     int errors = 0;
     int ordinal;
     int field_count_first;
@@ -287,10 +335,17 @@ check_rc(char *use_output)
     {
         if(use_output != NULL)
         {
+            free(s->output_name);
             s->output_name = xstrdup(use_output);
         }
 
+        if(s->output_name == NULL)
+        {
+            s->output_name = DEFAULT_OUTPUT;
+        }
+
         s->o = search_output(s->output_name);
+        s->max_record_len = 0;
         if(s->o == NULL) errors++;
         r = s->r;
         if(r == NULL) 
@@ -319,7 +374,7 @@ check_rc(char *use_output)
 
         while(r != NULL)
         {
-            if(r->output_name == NULL) 
+            if(r->output_name == NULL || use_output != NULL) 
             {
                 r->output_name = s->output_name;
                 r->o = s->o;
@@ -328,6 +383,13 @@ check_rc(char *use_output)
                 r->o = search_output(r->output_name);
                 if(r->o == NULL) errors++;
             }
+
+            if(several_records && s->type[0] == BINARY && r->i == NULL)
+            {
+                errors++;
+                fprintf(stderr,"%s: Every record in a binary multi-record structure must have an id, structure \'%s\', record \'%s\'\n",program,s->name,r->name);
+            }
+
             if(r->fields_from != NULL)
             {
                 if(r->f != NULL)
@@ -345,6 +407,7 @@ check_rc(char *use_output)
                     fprintf(stderr,"%s: No record named as '\%s\' in structure \'%s\'\n",program,r->fields_from,s->name);
                 }
             }
+
             f = r->f;
             if(f == NULL)
             {
@@ -355,9 +418,10 @@ check_rc(char *use_output)
             ordinal = 1;
             while(f != NULL) 
             {
-                if(s->type[0] == FIXED_LENGTH)
+                if(s->type[0] == FIXED_LENGTH || s->type[0] == BINARY)
                 {
                     f->position = r->length;
+                    f->bposition = r->length;
                     r->length += f->length;
                 } else
                 {
@@ -377,10 +441,23 @@ check_rc(char *use_output)
                     sprintf(num,"%d",ordinal);
                     f->name = xstrdup(num);
                 }
-                if(s->type[0] == FIXED_LENGTH && !f->length)
+
+                if(s->type[0] == BINARY && f->length < 1)
                 {
                     errors++;
-                    fprintf(stderr,"%s: A field \'%s\' must have length in fixed length structure \'%s\' record \'%s\'\n",program,f->name,s->name,r->name);
+                    fprintf(stderr,"%s: The field \'%s\' must have length in binary structure \'%s\' record \'%s\'\n",program ,f->name,s->name,r->name);
+                }
+
+                if(s->type[0] == FIXED_LENGTH && f->length < 1)
+                {
+                    if(f->next) /* last field can have length 0 */
+                    {
+                        errors++;
+                        fprintf(stderr,"%s: The field \'%s\' must have length in fixed length structure \'%s\' record \'%s\'\n",program,f->name,s->name,r->name);
+                    } else
+                    {
+                        r->arb_length = 1;
+                    }
                 }
 
                 if(f->lookup_table_name != NULL)
@@ -403,9 +480,23 @@ check_rc(char *use_output)
                     }
                 }
 
+                if(f->output_name != NULL && use_output == NULL)
+                {
+                    f->o = search_output(f->output_name);
+                    if(f->o == NULL) errors++;
+                    if(f->o == raw) {
+                        errors++;
+                        fprintf(stderr,"%s: Field cannot have output \'raw\', field \'%s\'\n",program,f->name);
+                    }
+                }
+
                 f = f->next;
                 ordinal++;
             }
+
+            if(r->length > s->max_record_len) s->max_record_len = r->length;
+            if(s->type[0] == BINARY && s->max_record_len > max_binary_record_length) max_binary_record_length = s->max_record_len;
+
             if(s->header && r->length != field_count_first)
             {
                 errors++;
@@ -464,10 +555,62 @@ add_replace(char *optarg)
 
     r->field = xstrdup(optarg);
     op_pos++;
-    r->value = xstrdup(op_pos);
+    r->value = (uint8_t *) xstrdup(op_pos);
     r->found = 0;
 }
 
+struct expr_list *
+read_expression_file(char *file)
+{
+    struct expr_list *retval,*e;
+    FILE *fp;
+    int ccount;
+    size_t line_len = 1024;
+    char *line = xmalloc(line_len);
+
+    e = NULL;
+    retval = NULL;
+
+    fp = xfopen(file,"r");
+
+    do
+    {
+#ifdef HAVE_GETLINE
+        ccount = getline(&line,&line_len,fp);
+#else
+        if(fgets(line,line_len,fp) != NULL)
+        {
+            ccount = strlen(line);
+        } else
+        {
+            ccount = -1;
+        }
+#endif
+        if (ccount != -1)
+        {
+            if(ccount) ccount--;
+            line[ccount] = 0;
+            if(e == NULL)
+            {
+                e = xmalloc(sizeof(struct expr_list));
+                e->value = xstrdup(line);
+                retval = e;
+            } else
+            {
+                e->next = xmalloc(sizeof(struct expr_list));;
+                e = e->next;
+                e->value = xstrdup(line);
+            }
+            e->value_len = ccount;
+            e->next = NULL;
+        }
+    }
+    while(ccount != -1);
+    fclose(fp);
+    free(line);
+    return retval;
+}
+
 
 void
 add_expression(char *optarg)
@@ -475,9 +618,11 @@ add_expression(char *optarg)
     char *op_pos;
     char op = 0;
     struct expression *e;
+    struct expr_list *el;
     int rc;
     size_t buflen;
     char *errbuf;
+    char *value_file;
 
     if((op_pos = strchr(optarg,OP_REQEXP)) != NULL)
     {
@@ -522,34 +667,120 @@ add_expression(char *optarg)
 
     e->field = xstrdup(optarg);
     op_pos++;
-    e->value = xstrdup(op_pos);
-    e->value_len = strlen(e->value);
+    if(strstr(op_pos,"file:") == op_pos)
+    {
+        value_file = expand_home(&op_pos[5]);
+        e->el = read_expression_file(value_file);
+        if(e->el == NULL) panic("No values in file",value_file,NULL);
+        free(value_file);
+    } else
+    {
+        e->value = xstrdup(op_pos);
+        e->value_len = strlen(e->value);
+        e->el = NULL;
+    }
     e->found = 0;
     e->op = op;
 #ifdef HAVE_REGEX
     if(e->op == OP_REQEXP)
     {
-        rc = regcomp(&e->reg,e->value,REG_EXTENDED | REG_NOSUB);
-        if(rc)
+        if(e->el != NULL)
         {
-            buflen = regerror(rc,&e->reg,NULL,0);
-            errbuf = xmalloc(buflen + 1);
-            regerror(rc,&e->reg,errbuf,buflen);
-            panic("Error in regular expression",e->value,errbuf);
+            el = e->el;
+            while(el != NULL)
+            {
+                rc = regcomp(&el->reg,el->value,REG_EXTENDED | REG_NOSUB);
+                if(rc)
+                {
+                    buflen = regerror(rc,&e->reg,NULL,0);
+                    errbuf = xmalloc(buflen + 1);
+                    regerror(rc,&e->reg,errbuf,buflen);
+                    panic("Error in regular expression",el->value,errbuf);
+                }
+                el = el->next;
+            }
+        } else
+        {
+            rc = regcomp(&e->reg,e->value,REG_EXTENDED | REG_NOSUB);
+            if(rc)
+            {
+                buflen = regerror(rc,&e->reg,NULL,0);
+                errbuf = xmalloc(buflen + 1);
+                regerror(rc,&e->reg,errbuf,buflen);
+                panic("Error in regular expression",e->value,errbuf);
+            }
         }
     }
 #endif
 }
-       
+
+/* if variable is set it will not be overwritten */
+void
+set_env(char *name, char *value)
+{
+        if(name == NULL || value == NULL) return;
+#ifdef HAVE_SETENV
+        if(setenv(name,value,0) != 0) problem("Cannot set environment",name,strerror(errno));
+#elif HAVE_PUTENV
+        if(getenv(name) == NULL)
+        {
+             char *s = xmalloc(strlen(name)+strlen(value)+2);
+             strcpy(s,name);
+             strcat(s,"=");
+             strcat(s,value);
+             if(putenv(s) != 0) problem("Cannot set environment",s,strerror(errno));
+        }
+#else
+        problem("Cannot set environment",NULL,NULL);
+#endif
+}
+
+
+
+/* set FFE_STRUCTURE,FFE_FORMAT,FFE_OUTPUT,FFE_FIRST_FILE,FFE_FILES,*/
+/* returns 0 in case of error */
+/* if variable is set it will not be overwritten */
+void
+environment(char *structure, char *format,char *ofile)
+{
+    struct input_file *f=files;
+    char *file_names;
+    size_t files_len = 1024,used = 0;
+    
+    set_env("FFE_STRUCTURE",structure);
+    set_env("FFE_OUTPUT",ofile);
+    set_env("FFE_FORMAT",format);
+    set_env("FFE_FIRST_FILE",f->name);
+    file_names = xmalloc(files_len);
+    file_names[0] = 0;
+    while(f != NULL)
+    {
+        used += strlen(f->name) + 1;
+        if(used > files_len) 
+        {
+            files_len *= 2;
+            file_names = xrealloc(file_names,files_len);
+        }
+        strcat(file_names,f->name);
+        if(f->next != NULL) strcat(file_names," ");
+        f = f->next;
+    }
+    set_env("FFE_FILES",file_names);
+    free(file_names);
+}
+
+    
 
 int
 main(int argc, char **argv)
 {
     int opt;
     int strict = 1;
+    int debug = 0;
+    int info = 0;
     int expression_and = 0;
     int expression_invert = 0;
-    struct structure *s;
+    struct structure *s = NULL;
     char *structure_to_use = NULL;
     char *output_to_use = NULL;
     char *config_to_use = NULL;
@@ -617,6 +848,9 @@ main(int argc, char **argv)
                 case 'a':
                     expression_and = 1;
                     break;
+                case 'd':
+                    debug = 1;
+                    break;
                 case 'v':
                     expression_invert = !expression_invert;
                     break;
@@ -631,6 +865,9 @@ main(int argc, char **argv)
                     print_version();
                     exit(EXIT_SUCCESS);
                     break;
+                case 'I':
+                    info = 1;
+                    break;
                 default:
                     usage(opt);
                     exit(EXIT_FAILURE);
@@ -648,29 +885,46 @@ main(int argc, char **argv)
     }
     
     if(config_to_use == NULL) config_to_use = get_default_rc_name();
+
+    environment(structure_to_use,output_to_use,ofile_to_use);
+    
+    system_endianess = check_system_endianess();
                 
     parserc(config_to_use,field_list);
 
     check_rc(output_to_use);
 
-    open_input_file();
-
-    if(structure_to_use == NULL)  
+    if(info)
     {
-        structure_to_use = guess_structure();
-        if(structure_to_use == NULL) panic("A structure cannot be guessed, use -s option",NULL,NULL);
-    } 
+        print_info();
+        exit(EXIT_SUCCESS);
+    }
 
-    s = find_structure(structure_to_use);
-
-    if(s == NULL)
+    if(structure_to_use == NULL)
     {
-        panic("No structure named as",structure_to_use,NULL);
+        open_input_file(BINARY);  // guess binary first
+        structure_to_use = guess_binary_structure();
+        if(structure_to_use == NULL)
+        {
+            structure_to_use = guess_structure();
+        }
+        if(structure_to_use == NULL) panic("Structure cannot be guessed, use -s option",NULL,NULL);
+        s = find_structure(structure_to_use);
+    } else
+    {
+        s = find_structure(structure_to_use);
+        if(s != NULL)
+        {
+            open_input_file(s->type[0]);
+        } else
+        {
+            panic("No structure named as",structure_to_use,NULL);
+        }
     }
 
     set_output_file(ofile_to_use);
 
-    execute(s,strict,expression_and,expression_invert);
+    execute(s,strict,expression_and,expression_invert,debug);
 
     close_output_file();
 
