@@ -20,7 +20,7 @@
  *
  */
 
-/* $Id: ffe.h,v 1.38 2007-05-21 16:23:32 timo Exp $ */
+/* $Id: ffe.h,v 1.61 2008-06-15 10:59:41 timo Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -51,6 +51,18 @@
 #include <unistd.h>
 #endif 
 
+#ifdef HAVE_FCNTL_H
+#include <fcntl.h>
+#endif
+
+#ifdef HAVE_INTTYPES_H
+#include <inttypes.h>
+#endif
+
+#ifdef HAVE_STDINT_H
+#include <stdint.h>
+#endif
+
 #include <stdio.h>
 
 #if defined(__MINGW32__)
@@ -73,7 +85,7 @@
 
 /* Types */
 
-#define DEFALT_OUTPUT "default"
+#define DEFAULT_OUTPUT "default"
 #define LEFT_JUSTIFY 1
 #define RIGHT_JUSTIFY 2
 
@@ -90,16 +102,16 @@ struct include_field {
 
 struct output {
     char *name;
-    char *file_header;
-    char *file_trailer;
-    char *header;
-    char *data;
-    char *lookup;
-    char *separator;
-    char *record_header;
-    char *record_trailer;
-    char justify;
-    char *indent;
+    uint8_t *file_header;
+    uint8_t *file_trailer;
+    uint8_t *header;
+    uint8_t *data;
+    uint8_t *lookup;
+    uint8_t *separator;
+    uint8_t *record_header;
+    uint8_t *record_trailer;
+    uint8_t justify;
+    uint8_t *indent;
     int no_data;
     char *empty_chars;
     int print_empty;
@@ -108,8 +120,8 @@ struct output {
 };
 
 struct lookup_data {
-    char *key;
-    char *value;
+    uint8_t *key;
+    uint8_t *value;
     int key_len;
     struct lookup_data *next;
 };
@@ -117,7 +129,7 @@ struct lookup_data {
 struct lookup {
     char *name;
     char type; 
-    char *default_value;
+    uint8_t *default_value;
     int max_key_len;
     struct lookup_data *data;
     struct lookup *next;
@@ -126,26 +138,22 @@ struct lookup {
 /* replace, field and value */
 struct replace {
     char *field;
-    char *value;
+    uint8_t *value;
     int found;
     struct replace *next;
 };
 
-
-/* Information for one field */
-struct field {
-    char *name;
-    char *const_data;
-    int position; /* first position = 0 byte position for fixed, first position = 1 for field number for separated */
-    int bposition; /* position in current input buffer */
-    int length;
-    int print;
-    char *lookup_table_name;
-    struct lookup *lookup;
-    struct replace *rep;  /* non NULL if value should be replaced */
-    struct field *next;
+/* values read from expression file */
+struct expr_list {
+    char *value;
+    int value_len;
+#if HAVE_REGEX
+    regex_t reg;
+#endif
+    struct expr_list *next;
 };
 
+
 /* search expression */
 struct expression {
     char *field;
@@ -157,13 +165,34 @@ struct expression {
 #endif
     int found;
     struct field *f;  /* pointer to field used in expression */
+    struct expr_list *el; /* values read from expression file */
     struct expression *next;
 };
 
+/* Information for one field */
+struct field {
+    char *name;
+    char *const_data;
+    int type;     /* field type ASC,CHAR,SHORT,.... */
+    int endianess; /* binary data endianess */
+    int position; /* first position = 0 byte position for fixed, first position = 1 for field number for separated */
+    int bposition; /* position in current input buffer */
+    int length;
+    int print;
+    char *lookup_table_name;
+    struct lookup *lookup;
+    struct replace *rep;  /* non NULL if value should be replaced */
+    struct expression *exp;
+    char *output_name;
+    struct output *o;
+    struct field *next;
+};
+
+
 /* contains pointer to fields which will be printed */
 struct print_field {
     struct field *f;
-    char *data;                 // data start position in output buffer;
+    uint8_t *data;                 // data start position in output buffer;
     int justify_length;
     int empty;                  // does the field contain only "empty" chars
     struct print_field *next;
@@ -171,7 +200,12 @@ struct print_field {
 
 struct id {
     int position;
-    char *key;
+    uint8_t *key;
+    int regexp; /* 1 or 0 for regexp */
+#if HAVE_REGEX
+    regex_t reg;
+#endif
+    int length;
     struct id *next;
 };
 
@@ -186,11 +220,14 @@ struct record {
     char *output_name;
     int vote;
     int length;
+    int arb_length; /* arbitrary length */
     struct record *next;
 };
 
+#define NO_READ '0'
 #define FIXED_LENGTH 'f'
 #define SEPARATED 's'
+#define BINARY 'b'
 
 /* header types of input files */
 #define HEADER_NO 0
@@ -199,7 +236,8 @@ struct record {
 
 struct structure {
     char *name;
-    char type[3]; /* [0] = f or s,[1] = separator, [2] = asterisk or 0 */
+    char type[3]; /* [0] = f,b or s,[1] = separator, [2] = asterisk or 0 */
+    int max_record_len;
     char quote;
     int header;
     char *output_name;
@@ -227,10 +265,29 @@ struct input_file {
 #define OP_NOT_EQUAL '!'
 #define OP_REQEXP '?'
 
+/* field types */
+#define F_ASC 1
+#define F_INT 3
+#define F_CHAR 4
+#define F_FLOAT 7
+#define F_DOUBLE 8
+#define F_BCD 9
+#define F_UINT 10
+#define F_HEX 11
+
+/* endianess */
+#define F_UNKNOWN_ENDIAN 0
+#define F_BIG_ENDIAN 1
+#define F_LITTLE_ENDIAN 3
+#define F_SYSTEM_ENDIAN 4
+
 /* function prototypes */
 extern void 
 panic(char *msg,char *info,char *syserror);
 
+extern void 
+problem(char *msg,char *info,char *syserror);
+
 extern void *
 xmalloc (size_t size);
 
@@ -259,7 +316,28 @@ extern void
 close_output_file();
 
 extern void 
-execute(struct structure *,int,int,int);
+execute(struct structure *,int,int,int,int);
+
+extern char *
+expand_home(char *);
+
+extern FILE *
+xfopen(char *, char *);
+
+extern FILE *
+xfopenb(char *, char *);
+
+extern uint8_t *
+endian_and_align(uint8_t *,int,int,int);
+
+extern int
+check_system_endianess();
+
+extern char *
+guess_binary_structure();
+
+extern void
+file_to_text(FILE *);
 
 /* global variables */
 extern struct structure *structure;
@@ -267,4 +345,7 @@ extern struct output *output;
 extern struct expression *expression;
 extern struct lookup *lookup;
 extern struct output *no_output;
+extern struct output *raw;
 extern struct field *const_field;
+extern int system_endianess;
+extern int max_binary_record_length;
