@@ -20,7 +20,7 @@
  *
  */
 
-/* $Id: parserc.c,v 1.41 2007-05-30 07:32:48 timo Exp $ */
+/* $Id: parserc.c,v 1.76 2008-09-11 06:44:21 timo Exp $ */
 /* parsing the rc-file */
 
 #include "ffe.h"
@@ -28,6 +28,15 @@
 #include <string.h>
 #include <ctype.h>
 
+#ifdef HAVE_PROCESS_H
+#include <process.h>
+#endif
+
+#ifdef HAVE_IO_H
+#include <io.h>
+#endif
+
+
 #ifdef PACKAGE
 static char *program = PACKAGE;
 #else
@@ -36,6 +45,7 @@ static char *program = "ffe";
 
 /* pointer to rc-file */
 static FILE *fp = NULL;
+static FILE *child = NULL;      /* read from child */
 
 /* current line number */
 static int lineno = 0;
@@ -52,6 +62,11 @@ static int lineno = 0;
 #define LL_BLOCK_END 3
 #define LL_EOF 4
 
+#define COMMAND_LEN 128
+static size_t command_len = COMMAND_LEN;
+static char *command;                    /* used in command substituion */
+static char *cs_temp_file = NULL;
+
 /* structure containing one rc-file option definition */
 struct rc_option {
     char *name;              /* the name */
@@ -79,6 +94,7 @@ char *values[100];            /* poister to option name and parameters */
 #define N_OUTPUT            "output"
 #define N_RECORD            "record"
 #define N_ID                "id"
+#define N_RID               "rid"
 #define N_FIELD             "field"
 #define N_FIELDSFROM        "fields-from"
 #define N_FILE_HEADER       "file-header"
@@ -99,6 +115,7 @@ char *values[100];            /* poister to option name and parameters */
 #define N_DEFAULT           "default-value"
 #define N_SEARCH            "search"
 #define N_CONST             "const"
+#define N_FIELD_COUNT       "field-count"
 
 
 
@@ -111,7 +128,8 @@ static struct rc_option rc_opts[] = {
     {N_RECORD,"S"},
     {N_OUTPUT,"S"},
     {N_ID,"NS"},
-    {N_FIELD,"Sns"},
+    {N_RID,"NS"},
+    {N_FIELD,"Ssss"},
     {N_FIELDSFROM,"S"},
     {N_FILE_HEADER,"S"},
     {N_FILE_TRAILER,"S"},
@@ -131,15 +149,33 @@ static struct rc_option rc_opts[] = {
     {N_DEFAULT,"S"},
     {N_SEARCH,"S"},
     {N_CONST,"SS"},
+    {N_FIELD_COUNT,"N"},
     {NULL,NULL}
 };
 
+int
+is_digit(char *number)
+{
+    while(*number) if(!isdigit(*number++)) return 0;
+    return 1;
+}
+
+void remove_temp_file()
+{
+    if(cs_temp_file) unlink(cs_temp_file);
+}
+
 
 void
 open_rc_file(char *file)
 {
-    fp = fopen(file,"r");
-    if (fp == NULL) panic("Error in opening file",file,strerror(errno));
+    fp = xfopen(file,"r");
+}
+
+void 
+error_in_line()
+{
+    fprintf(stderr,"%s: Error in rcfile, line %d\n",program,lineno);
 }
 
 /* remove leading and trailing whitespace */
@@ -227,16 +263,139 @@ parse_include_list(char *list)
     }
     return ret;
 }
+
+/* start child for command substitution */
+/* returns a stream for reading from child */
+FILE *
+execute_child(char *command)
+{
+    int fds[2];
+    pid_t pid;
+    FILE *ret = NULL;
+#if defined(HAVE_WORKING_FORK) && defined(HAVE_DUP2) && defined(HAVE_PIPE)
+    if (pipe(fds) != 0) panic("Cannot create pipe",strerror(errno),NULL);
+    pid = fork();
+    if(pid == (pid_t) 0) /* Child */
+    {
+        close(fds[0]);
+        if(dup2(fds[1],STDOUT_FILENO) == -1) panic("dup2 error",strerror(errno),NULL);
+        if(system(command) == -1) panic("Command failed",command,strerror(errno));
+        close(fds[1]);
+        _exit(EXIT_SUCCESS);
+    } else if(pid > (pid_t) 0)
+    {
+        close(fds[1]);
+        ret = fdopen(fds[0],"r");
+        if(ret == NULL) panic("Cannot read from command",command,strerror(errno));
+    } else
+    {
+        panic("Cannot fork",strerror(errno),NULL);
+    }
+#elif defined(HAVE_TEMPNAM)  /* for win etc. */
+    cs_temp_file = tempnam(NULL,"ffe");
+    if(tmpfile == NULL) panic("Temporary file cannot be created",strerror(errno),NULL);
+    if((size_t) (strlen(command) + 4 + strlen(cs_temp_file)) > command_len)
+    {
+        command_len = (size_t) (strlen(command) + 4 + strlen(cs_temp_file));
+        command = xrealloc(command,command_len);
+    }
+    strcat(command," > ");
+    strcat(command,cs_temp_file);
+    if(system(command) == -1) panic("Command failed",command,strerror(errno));
+    ret = fopen(cs_temp_file,"r");
+    if(ret == NULL) panic("Temporary file cannot be read",cs_temp_file,strerror(errno));
+#else
+    panic("Command substitution is not supported in this system",NULL,NULL);
+#endif
+#if defined(HAVE_SETMODE) && defined(WIN32)
+    setmode(fileno(ret),O_TEXT);
+#endif
+    return ret;
+}
+
+
+
+/* reads a chracter from the rc-file. */
+/* if command substition is used, char is read from child process */
+int
+read_char()
+{
+    int c;
+    static int char_in_buffer = 0; /* contains the "peeked" char after newline */
+
+    if(child != NULL) 
+    {
+        if(char_in_buffer)
+        {
+            c = char_in_buffer;
+            char_in_buffer = 0;
+        } else
+        {
+            c = getc(child);
+            if(c == '\n')  /* check trailing newlines */
+            {
+                while((c = getc(child)) == '\n');
+                if(c != EOF)      /* abandon trailing newlines */
+                {
+                    char_in_buffer = c; /* there is more lines, put c in safe place, return a single newline */
+                    c = '\n';
+                } 
+            }  /* No else ! */
+
+            if(c == EOF) 
+            {
+                fclose(child);
+                if(cs_temp_file) unlink(cs_temp_file);
+                cs_temp_file = NULL;
+                child = NULL;
+                c = read_char();
+            }
+        }
+    } else
+    {
+        c = getc(fp);
+        if(c == '`')  /* command subst. starts */
+        {
+            size_t i=0;
+            do
+            {
+                c = getc(fp);
+                command[i++] = c;
+                if(i == command_len) 
+                {
+                    command_len = 2*command_len;
+                    command = xrealloc(command,command_len);
+                }
+            } while(c != '`' && c != EOF);
+            if(c != '`')
+            {
+                error_in_line();
+                panic("No closing '`' in command substition",NULL,NULL);
+            }
+            command[--i] = 0;
+            child = execute_child(command);
+            c = read_char();
+        }
+    }
+    return c;
+}
+
+            
+
         
 /* reading one logical line, returns the status of the line read */
 int 
-read_logical_line(char *buffer, int bufsize)
+read_logical_line(char **buffer,size_t *bufsize)
 {
     static char last_eol_char = 0;      /* for what did previous read end */
-    register int prev_char = 0,c;
-    register int i=0,retval = 0;
+    register char *bp;
+    int prev_char = 0,c;
+    register int i=0;
+    int retval = 0;
 
-    buffer[0] = 0;
+    bp = *buffer;
+
+    bp[0] = 0;
 
     switch(last_eol_char)
     {
@@ -252,64 +411,64 @@ read_logical_line(char *buffer, int bufsize)
     
     do
     {
-        c = getc(fp);
+        c = read_char();
         if(prev_char == '\\')
         {
             switch(c)
             {
                 case 'a':
-                    buffer[i] = '\a';
+                    bp[i] = '\a';
                     c = 0;
                     i++;
                     break;
                 case 'b':
-                    buffer[i] = '\b';
+                    bp[i] = '\b';
                     c = 0;
                     i++;
                     break;
                 case 't':
-                    buffer[i] = '\t';
+                    bp[i] = '\t';
                     c = 0;
                     i++;
                     break;
                 case 'n':
-                    buffer[i] = '\n';
+                    bp[i] = '\n';
                     c = 0;
                     i++;
                     break;
                 case 'v':
-                    buffer[i] = '\v';
+                    bp[i] = '\v';
                     c = 0;
                     i++;
                     break;
                 case 'f':
-                    buffer[i] = '\f';
+                    bp[i] = '\f';
                     c = 0;
                     i++;
                     break;
                 case 'r':
-                    buffer[i] = '\r';
+                    bp[i] = '\r';
                     c = 0;
                     i++;
                     break;
                 case '\\':
-                    buffer[i] = '\\';
+                    bp[i] = '\\';
                     c = 0;
                     i++;
                     break;
                 case COMMENT:
-                    buffer[i] = COMMENT;
+                    bp[i] = COMMENT;
                     c = 0;
                     i++;
                     break;
                 case '\n':    /* newline escaped */
-                    lineno++;
-                    buffer[i] = getc(fp);
+                    if(child == NULL) lineno++; 
+                    bp[i] = read_char();
                     c = 0;
                     i++;
                     break;
                 default:
-                    buffer[i] = '\\';
+                    bp[i] = '\\';
                     i++;
                     break;
             }
@@ -322,23 +481,23 @@ read_logical_line(char *buffer, int bufsize)
             case COMMENT:
                 do
                 {
-                    c = getc(fp);
+                    c = read_char();
                 } while (c != '\n' && c != EOF);  /* no break !*/
             case '\n':
-                if(c == '\n') lineno++;
-                buffer[i] = 0;
+                if(c == '\n' && child == NULL) lineno++;
+                bp[i] = 0;
                 retval = LL_OPTION;
                 break;
             case BLOCK_START:
-                buffer[i] = 0;
+                bp[i] = 0;
                 retval = LL_OPTION;
                 break;
             case BLOCK_END:
-                buffer[i] = 0;
+                bp[i] = 0;
                 retval = LL_OPTION;
                 break;
             case EOF:
-                buffer[i] = 0;
+                bp[i] = 0;
                 retval = LL_OPTION;
                 break;
             case '\\':
@@ -346,19 +505,24 @@ read_logical_line(char *buffer, int bufsize)
             case 0:
                 break;
             default:
-                if(i >= bufsize) panic("rc-file line too long",NULL,NULL);
-                buffer[i] = (char) c;
+                if(i >= *bufsize - 10) 
+                {
+                    *bufsize *= 2;
+                    *buffer = xrealloc(*buffer,*bufsize);
+                    bp = *buffer;
+                }
+                bp[i] = (char) c;
                 i++;
                 break;
         }
 
         if(retval)
         {
-            trim(buffer);
+            trim(bp);
             switch(c)
             {
                 case '\n':
-                    if(buffer[0] == 0) 
+                    if(bp[0] == 0) 
                     {
                         i = 0;          /* empty line */
                         retval = 0;
@@ -368,13 +532,13 @@ read_logical_line(char *buffer, int bufsize)
                     }
                     break;
                 case EOF:
-                    if(buffer[0] == 0)
+                    if(bp[0] == 0)
                     {
                          retval = LL_EOF;
                     }
                     return retval;
                 case BLOCK_START:
-                    if(buffer[0] == 0)
+                    if(bp[0] == 0)
                     {
                         retval = LL_BLOCK_START;
                     } else
@@ -383,7 +547,7 @@ read_logical_line(char *buffer, int bufsize)
                     }
                     return retval;
                 case BLOCK_END:
-                    if(buffer[0] == 0)
+                    if(bp[0] == 0)
                     {
                         retval = LL_BLOCK_END;
                     } else
@@ -397,12 +561,6 @@ read_logical_line(char *buffer, int bufsize)
     } while (1);
 }
 
-void 
-error_in_line()
-{
-    fprintf(stderr,"%s: Error in rcfile, line %d\n",program,lineno);
-}
-
 #define READ_BUF_SIZE 1024
 
 /* reads logical lines and parses a one option */
@@ -572,7 +730,7 @@ int parse_option(char *buf)
                         }
                         break;
                 }
-                if(valc > 3)
+                if(valc > 4)
                 {
                     error_in_line();
                     panic("Too many parameters",values[0],NULL);
@@ -632,13 +790,8 @@ read_lookup_from_file(struct lookup_data **data,char *file,char separator,int *m
 
     efile = expand_home(file);
     
-    fp = fopen(efile,"r");
+    fp = xfopen(efile,"r");
 
-    if(fp == NULL) {
-        error_in_line();
-        panic("Cannot open file",efile,strerror(errno));
-    }
-    
     line = xmalloc(max_line_size);
 
     if(c_data != NULL) while(c_data->next != NULL) c_data = c_data->next;
@@ -696,7 +849,227 @@ read_lookup_from_file(struct lookup_data **data,char *file,char separator,int *m
     return c_data;
 }
 
-                           
+/* non printable characters in form \xnn will be expanded 
+   returns id length
+*/
+int
+expand_non_print(char *s,uint8_t **t)
+{
+    char num[5];
+    char *w;
+    int c,len;
+
+    num[0] = '0';
+    num[1] = 'x';
+    num[4] = 0;
+    len = 0;
+    
+    *t = xstrdup(s);
+    w = *t;
+
+    while(*s)
+    {
+        if(*s == '\\' && s[1] == 'x' && isxdigit((int) s[2]) && isxdigit((int) s[3]))
+        {
+            num[2] = s[2];
+            num[3] = s[3];
+            sscanf(num,"%i",&c);
+            *w = (char) c;
+            s += 3;
+        } else
+        {
+            *w = *s;
+        }
+        s++;
+        w++;
+        len++;
+    }
+    *w = 0;
+    return len;
+}
+
+/* Parse non numeric field type */
+    void
+parse_field_type(char *t,struct field *f)
+{
+    if(strcmp(t,"char") == 0)
+    {
+        f->type = F_CHAR;
+        f->length = sizeof(char);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"short") == 0)
+    {
+        f->type = F_INT;
+        f->length = sizeof(short int);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"int") == 0) 
+    {
+        f->type = F_INT;
+        f->length = sizeof(int);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"long") == 0) 
+    {
+        f->type = F_INT;
+        f->length = sizeof(long int);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"llong") == 0) 
+    {
+        f->type = F_INT;
+        f->length = sizeof(long long int);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"ushort") == 0)
+    {
+        f->type = F_UINT;
+        f->length = sizeof(unsigned short int);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"uint") == 0) 
+    {
+        f->type = F_UINT;
+        f->length = sizeof(unsigned int);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"ulong") == 0) 
+    {
+        f->type = F_UINT;
+        f->length = sizeof(unsigned long int);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"ullong") == 0) 
+    {
+        f->type = F_UINT;
+        f->length = sizeof(long long int);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"float") == 0) 
+    {
+        f->type = F_FLOAT;
+        f->length = sizeof(float);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"float_be") == 0) 
+    {
+        f->type = F_FLOAT;
+        f->length = sizeof(float);
+        f->endianess = F_BIG_ENDIAN;
+    } else if(strcmp(t,"float_le") == 0) 
+    {
+        f->type = F_FLOAT;
+        f->length = sizeof(float);
+        f->endianess = F_LITTLE_ENDIAN;
+    } else if(strcmp(t,"double") == 0) 
+    {
+        f->type = F_DOUBLE;
+        f->length = sizeof(double);
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"double_be") == 0) 
+    {
+        f->type = F_DOUBLE;
+        f->length = sizeof(double);
+        f->endianess = F_BIG_ENDIAN;
+    } else if(strcmp(t,"double_le") == 0) 
+    {
+        f->type = F_DOUBLE;
+        f->length = sizeof(double);
+        f->endianess = F_LITTLE_ENDIAN;
+    } else if(strcmp(t,"int8") == 0)
+    {
+        f->type = F_INT;
+        f->length = 1;
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"int16_be") == 0) 
+    {
+        f->type = F_INT;
+        f->length = 2;
+        f->endianess = F_BIG_ENDIAN;
+    } else if(strcmp(t,"int16_le") == 0) 
+    {
+        f->type = F_INT;
+        f->length = 2;
+        f->endianess = F_LITTLE_ENDIAN;
+    } else if(strcmp(t,"int32_be") == 0) 
+    {
+        f->type = F_INT;
+        f->length = 4;
+        f->endianess = F_BIG_ENDIAN;
+    } else if(strcmp(t,"int32_le") == 0) 
+    {
+        f->type = F_INT;
+        f->length = 4;
+        f->endianess = F_LITTLE_ENDIAN;
+    } else if(strcmp(t,"int64_be") == 0) 
+    {
+        f->type = F_INT;
+        f->length = 8;
+        f->endianess = F_BIG_ENDIAN;
+    } else if(strcmp(t,"int64_le") == 0) 
+    {
+        f->type = F_INT;
+        f->length = 8;
+        f->endianess = F_LITTLE_ENDIAN;
+    } else if(strcmp(t,"uint8") == 0)
+    {
+        f->type = F_UINT;
+        f->length = 1;
+        f->endianess = system_endianess;
+    } else if(strcmp(t,"uint16_be") == 0) 
+    {
+        f->type = F_UINT;
+        f->length = 2;
+        f->endianess = F_BIG_ENDIAN;
+    } else if(strcmp(t,"uint16_le") == 0) 
+    {
+        f->type = F_UINT;
+        f->length = 2;
+        f->endianess = F_LITTLE_ENDIAN;
+    } else if(strcmp(t,"uint32_be") == 0) 
+    {
+        f->type = F_UINT;
+        f->length = 4;
+        f->endianess = F_BIG_ENDIAN;
+    } else if(strcmp(t,"uint32_le") == 0) 
+    {
+        f->type = F_UINT;
+        f->length = 4;
+        f->endianess = F_LITTLE_ENDIAN;
+    } else if(strcmp(t,"uint64_be") == 0) 
+    {
+        f->type = F_UINT;
+        f->length = 8;
+        f->endianess = F_BIG_ENDIAN;
+    } else if(strcmp(t,"uint64_le") == 0) 
+    {
+        f->type = F_UINT;
+        f->length = 8;
+        f->endianess = F_LITTLE_ENDIAN;
+    } else if(strncmp(t,"bcd_be_",7) == 0) 
+    {
+        f->type = F_BCD;
+        if(is_digit(&t[7])) sscanf(&t[7],"%i",&f->length);
+        f->endianess = F_BIG_ENDIAN;
+    } else if(strncmp(t,"bcd_le_",7) == 0) 
+    {
+        f->type = F_BCD;
+        if(is_digit(&t[7])) sscanf(&t[7],"%i",&f->length);
+        f->endianess = F_LITTLE_ENDIAN;
+    } else if (strncmp(t,"hex_be_",7) == 0)
+    {
+        f->type = F_HEX;
+        if(is_digit(&t[7])) sscanf(&t[7],"%i",&f->length);
+        f->endianess = F_BIG_ENDIAN;
+    } else if (strncmp(t,"hex_le_",7) == 0)
+    {
+        f->type = F_HEX;
+        if(is_digit(&t[7])) sscanf(&t[7],"%i",&f->length);
+        f->endianess = F_LITTLE_ENDIAN;
+    } else
+    {
+        error_in_line();
+        panic("Unknown field type",t,NULL);
+    }
+
+    if(f->length < 1)
+    {
+        error_in_line();
+        panic("Error in field type",t,NULL);
+    }
+}
+
 /* parse status values */
 #define PS_MAIN 1
 #define PS_STRUCT 2
@@ -709,28 +1082,54 @@ read_lookup_from_file(struct lookup_data **data,char *file,char separator,int *m
 #define PS_W_LOOKUP 9
 
 void
-print_s()
+print_info()
 {
     struct structure *s = structure;
     struct record *r;
     struct field *f;
+    int pos;
 
     while(s != NULL)
     {
-        printf("%s\n",s->name);
+        printf("Structure %s - ",s->name);
+        switch(s->type[0])
+        {
+            case BINARY:
+                printf("binary");
+                break;
+            case FIXED_LENGTH:
+                printf("fixed length");
+                break;
+            case SEPARATED:
+                printf("separated by '%c'",(int) s->type[1]);
+                break;
+        }
+        printf(" - %d\n",s->max_record_len);
+
         r = s->r;
         while(r != NULL)
         {
             f = r->f;
-            printf(" %s\n",r->name);
+            pos = 1;
+            printf("  Record %s  %d\n",r->name,r->length);
             while(f != NULL)
             {
-                printf("  %s %d\n",f->name,f->length);
+                if(s->type[0] == SEPARATED)
+                {
+                    printf("    Field %-30s%5d%5d\n",f->name == NULL ? "*" : f->name,pos,f->length);
+                    pos++;
+                } else
+                {
+                    printf("    Field %-30s%5d%5d\n",f->name,pos,f->length);
+                    pos += f->length;
+                }
                 f = f->next;
             }
             r = r->next;
+            printf("\n");
         }
         s = s->next;
+        printf("\n\n");
     }
 }
 
@@ -747,19 +1146,27 @@ parserc(char *rcfile,char *include_field_list)
     struct include_field *fl = parse_include_list(include_field_list);
 
     char *read_buffer;
+    size_t read_buffer_size;
     int line_status;
     int status = PS_MAIN;
     int opt_count;
+    int field_count;
 
     open_rc_file(rcfile);
 
-    read_buffer = xmalloc(READ_BUF_SIZE);
+    read_buffer_size = READ_BUF_SIZE;
+    read_buffer = xmalloc(read_buffer_size);
+    command = xmalloc(command_len);         /* used in command substituion */
+
+#ifdef HAVE_ATEXIT
+    atexit(remove_temp_file); /* remove possible temp-file */
+#endif
 
     if(c_structure != NULL) while(c_structure->next != NULL) c_structure = c_structure->next;
 
     if(c_output != NULL) while(c_output->next != NULL) c_output = c_output->next;
 
-    while((line_status = read_logical_line(read_buffer,READ_BUF_SIZE)) != LL_EOF)
+    while((line_status = read_logical_line(&read_buffer,&read_buffer_size)) != LL_EOF)
     {
         switch(line_status)
         {
@@ -784,7 +1191,7 @@ parserc(char *rcfile,char *include_field_list)
                             c_structure->type[0] = FIXED_LENGTH;
                             c_structure->quote = 0;
                             c_structure->header = 0;
-                            c_structure->output_name = DEFALT_OUTPUT;
+                            c_structure->output_name = NULL;
                             c_structure->vote = 0;
                             c_structure->o = NULL;
                             c_structure->r = NULL;
@@ -862,6 +1269,7 @@ parserc(char *rcfile,char *include_field_list)
                             c_field->const_data = xstrdup(values[2]);
                             c_field->position = 0;
                             c_field->length = strlen(c_field->const_data);
+                            c_field->o = NULL;
                         } else 
                         {
                             error_in_line();
@@ -900,6 +1308,14 @@ parserc(char *rcfile,char *include_field_list)
                                         panic("An \'*\' is expected",values[0],NULL);
                                     }
                                 }
+                            } else if(strcmp(values[1],"binary") == 0)
+                            {
+                                c_structure->type[0] = BINARY;
+                                if(opt_count > 1)
+                                {
+                                    error_in_line();
+                                    panic("too many parameters for",values[0],NULL);
+                                }
                             } else
                             {
                                 error_in_line();
@@ -943,6 +1359,7 @@ parserc(char *rcfile,char *include_field_list)
                             c_record->o = NULL;
                             c_record->output_name = NULL;
                             c_record->vote = 0;
+                            c_record->arb_length = 0;
                             status = PS_W_RECORD;
                         } else if(strcmp(values[0],N_QUOTE) == 0)
                         {
@@ -960,7 +1377,7 @@ parserc(char *rcfile,char *include_field_list)
                         }
                         break;
                     case PS_RECORD:
-                        if(strcmp(values[0],N_ID) == 0)
+                        if(strcmp(values[0],N_ID) == 0 || strcmp(values[0],N_RID) == 0)
                         {
                             if(c_record->i == NULL)
                             {
@@ -972,6 +1389,7 @@ parserc(char *rcfile,char *include_field_list)
                                 c_id = c_id->next;
                             }
                             c_id->next = NULL;
+                            c_id->regexp = 0;
                             if(sscanf(values[1],"%d",&c_id->position) != 1)
                             {
                                 error_in_line();
@@ -982,7 +1400,28 @@ parserc(char *rcfile,char *include_field_list)
                                 error_in_line();
                                 panic("Position must be greater than zero",NULL,NULL);
                             }
-                            c_id->key = xstrdup(values[2]);
+                            c_id->length = expand_non_print(values[2],&c_id->key);
+                            if(strcmp(values[0],N_RID) == 0)
+                            {
+#ifdef HAVE_REGEX
+                                int rc,buflen;
+                                char *errbuf;
+
+                                rc = regcomp(&c_id->reg,c_id->key,REG_EXTENDED | REG_NOSUB);
+                                if(rc)
+                                {
+                                    buflen = regerror(rc,&c_id->reg,NULL,0);
+                                    errbuf = xmalloc(buflen + 1);
+                                    regerror(rc,&c_id->reg,errbuf,buflen);
+                                    error_in_line();
+                                    panic("Error in regular expression",c_id->key,errbuf);
+                                }
+                                c_id->regexp = 1;
+#else
+                                error_in_line();
+                                panic("Regular expressions are not supported in this system",NULL,NULL);
+#endif
+                            }
                         } else if(strcmp(values[0],N_FIELD) == 0)
                         {
                             if(c_record->f == NULL)
@@ -997,9 +1436,14 @@ parserc(char *rcfile,char *include_field_list)
                             c_field->lookup_table_name = NULL;
                             c_field->lookup = NULL;
                             c_field->rep = NULL;
+                            c_field->type = F_ASC;
                             c_field->next = NULL;
                             c_field->const_data = NULL;
+                            c_field->exp = NULL;
                             c_field->length = 0;
+                            c_field->output_name = NULL;
+                            c_field->o = NULL;
+
                             if(values[1][0] == '*' && !values[1][1])
                             {
                                 c_field->name = NULL;
@@ -1011,15 +1455,24 @@ parserc(char *rcfile,char *include_field_list)
                             {
                                 if(!(values[2][0] == '*' && !values[2][1]))
                                 {
-                                    if(sscanf(values[2],"%d",&c_field->length) != 1)
+                                    if(is_digit(values[2]))
                                     {
-                                        error_in_line();
-                                        panic("Error in number",NULL,NULL);
+                                        sscanf(values[2],"%d",&c_field->length);
+                                    } else
+                                    {
+                                        parse_field_type(values[2],c_field);
                                     }
                                 }
                                 if(opt_count > 2)
                                 {
-                                    c_field->lookup_table_name = xstrdup(values[3]);
+                                    if(!(values[3][0] == '*' && !values[3][1]))
+                                    {
+                                        c_field->lookup_table_name = xstrdup(values[3]);
+                                    }
+                                    if(opt_count > 3)
+                                    {
+                                        c_field->output_name = xstrdup(values[4]);
+                                    }
                                 }
                             } 
                         } else if(strcmp(values[0],N_FIELDSFROM) == 0)
@@ -1028,6 +1481,31 @@ parserc(char *rcfile,char *include_field_list)
                         } else if(strcmp(values[0],N_OUTPUT) == 0)
                         {
                             c_record->output_name = xstrdup(values[1]);
+                        } else if(strcmp(values[0],N_FIELD_COUNT) == 0)
+                        {
+                            field_count = atoi(values[1]);
+                            while(field_count--)
+                            {
+                                if(c_record->f == NULL)
+                                {
+                                    c_field = xmalloc(sizeof(struct field));
+                                    c_record->f = c_field;
+                                } else
+                                {
+                                    c_field->next = xmalloc(sizeof(struct field));
+                                    c_field = c_field->next;
+                                }
+                                c_field->lookup_table_name = NULL;
+                                c_field->lookup = NULL;
+                                c_field->rep = NULL;
+                                c_field->next = NULL;
+                                c_field->const_data = NULL;
+                                c_field->exp = NULL;
+                                c_field->length = 0;
+                                c_field->name = NULL;
+                                c_field->output_name = NULL;
+                                c_field->o = NULL;
+                            }
                         } else
                         {
                             error_in_line();
